% ------------------------------------------
% formGen 0.1.0 29 August 2021
% free to use but no warranties
% El Condor - Condor Informatique - Turin
% ------------------------------------------
classdef formGen
    properties
        xPos = 10;                  % label position
        deltaY = 30;                % control's distance
        charWidth = 7;              % medium pixel per character
        buttonsDim = 70;            % button's dimension
        lineTerminator = newline;   % for text area
        checkBoxValues = {'On','Off'};
        gfc;                        % figure
        windowStyle = 'normal';     % normal|modal
    end
    properties(Constant)
      reInteger = '^[1-9]\d*$|^0$'; %   ^0$|^[1-9]\d*$
      reSignedInteger = '^[+-]?[1-9]\d*$|^0$';
      reFloating = '^[+-]?[1-9]\d*(\.\d+)?$|^0$';
    end
	methods (Static = true)
      function value = valueOf(this,name)
        h = guidata(this.gfc);
        data = h.data.(name);
        type = data{1};
        switch (type)     
            case "A"
              if (exist('OCTAVE_VERSION','builtin'))
                value = get(data{9},"string");
                if iscell(value)
                  value = strjoin(value,this.lineTerminator);
                end
              else
                value = get(data{9},"string");
                [row, column] = size(value);
                for i = 1:row                    
                    t{i} = strip(value(i,:),"Right");
                end
                value = strjoin(t,this.lineTerminator);
              end               
            case "CKB"
              uData = get(data{9},"userdata");
              value = cell2mat(uData(2 - get(data{9},"value")));
            case "H"
              value = data{3};
            case "I"
              value = str2double(strcat(get(data{9},"string"),"+",get(data{10},"string"),"i"));
            case {"L"}
               value = char(h.itemsList.(name)(get(data{9},"value")));
            case {"N","NS","NF"}
              value = str2double(get(data{9},"string"));
            case "R"
               rdb = get(data{9},"selectedobject");
               value = char(h.itemsList.(name)(str2double(get(rdb,"userdata"))));
            case "S"
               value = get(data{9},"value");
            otherwise
               value = get(data{9},'string');
        end
      end
      function res = fg_Inside(a,n1,n2)
          res = 0;
          if (a >= min(n1,n2))
              res = a <= max(n1,n2);
          end
      end
      function tooltip(handle,tip)
        if not(isempty(tip))
            set(handle,'tooltip',tip);
        end
      end
      function goTry()
        form = "FORM,,Form deGenerator;L,Unit,Measure Unit,,,mm:millimeter|cm:centimeter|m:meter|km:kilometer;I,iNumber,Imaginary number";
        formGen.fGen(formGen,form);
      end
      function [res, plcHolders] = mask(form)   % mask item that contain comma or semicolon
        rg = "[,]?(\'(.*?)\'|""(.*?)"")[,;]";
        x = char(form);
        [~, ~, p] = regexp(x,rg);
        y = '';
        startPoint = 1;
        plcHolders = struct();
        for i = 1:numel(p)
            z=cell2mat(p(i));
            plcholder = sprintf("@%i",i);
            if (startPoint == z(1,1)) y = strcat(y,plcholder);
            else y = strcat(y,x(startPoint:z(1,1)-1),plcholder);
            end
            plcHolders.(sprintf("ph%i",i)) = x(z(1,1)+1:z(1,2)-1);
            startPoint = z(1,2) + 1;
        end
        res = strcat(y,x(startPoint:numel(x)));
      end
      function res = replace(item,replStruct)
        reInteger = '^%(\d{2,5})$';
        res = item;
        if (numel(item) == 0) return; end
        switch(item(1:1))
            case "@"
                key = sprintf("ph%s",(item(2:numel(item)))); 
                if (isfield(replStruct,key))
                    res = replStruct.(key);
                end
            case "%"
                [~, ~, p] = regexp(item,reInteger);
                z = cell2mat(p(1));
                res = char(str2double(item(2:z(2))));                
        end
      end
      function res = fg_Left(str,n)
          res = str(1:min(n,numel(str)));
      end
      function posiz = setPos(posiz,w,n)
          persistent where
          where = struct("left",1,"bottom",2,"width",3,"height",4);
          if isfield(where,lower(w))
              posiz(where.(lower(w))) = n;
          end
      end
      function color = checkColor(color,defColor)
         persistent re
         re = "^#[0-9a-fA-F]{3}$";
         persistent colors
         colors = struct("red",0, "green",0, "blue",0, "cyan",0, "magenta",0, "yellow",0, "black",0, "white",0);
         answ = regexp(color,re);
         if numel(answ) > 0
           color = [hex2dec(color(2))/15 hex2dec(color(3))/15 hex2dec(color(4))/15];
         else
           if ~isfield(colors,lower(color))
             sprintf("Unknown color: %s",color)
             color = defColor;
           end
         end
      end
      function [k, v] = getKeyValue (l)
          lItems = strsplit (l, '|','CollapseDelimiters',false);
          k = cell(1,numel(lItems));
          v = cell(1,numel(lItems));
          for i=1:numel(lItems)
              a = strsplit(char(strcat(lItems(i),':',lItems(i))),':');
              k(i) = a(1);
              v(i) = a(2);
          end
      end
      function realWidth = checkWidth(realWidth, w)
        if realWidth < w(1) + w(3) realWidth =  w(1) + w(3); end
      end
    function p = fGen (this,params,handleAnswer)
        if (nargin < 3)
            sprintf("\nUse:\n\tobj = formGen;\n\tobj.fGen(formGen,controlsList,@handleFunction);\nor\n\tformGen.fGen(formGen,,controlsList,@handleFunction)")
            return
        end
%       params
        xPos = this.xPos;
        yPos = 36;              % bottom 
        deltaY = this.deltaY;
        hWidget = 0.8*deltaY;
        charWidth = this.charWidth;
        txtWidth = 50;      % 1 line textx width
        buttonsDim = this.buttonsDim;
        lLabels = 0;
        nWidget = 0;
        realWidth = 0;      % for resize window
        buttons.fg_Ok = {@fg_exit ["Ok"]};
        buttons.fg_Cancel = {@fg_exit ["Cancel"]};
        buttons.fg_Reset = {@fg_reset ["Reset"]};
        defLength = struct("A","100","T","30","I","9","N","7","NS","8","NF","9","R","12");
        numerics = struct("I",0,"N",0, "NS",0, "NF", 0);
%       ***** form parameters *****
        Title = "";
        formColor =  [0.94 0.94 0.94];   % same for labels and groups
        formPosition = [-1 -1 -1 -1];         % if present
        Links = struct();          % fields linked structure
        gData.Events = struct();   % events structure
        gData.data = struct();  
        [params, plcHolders] = this.mask(params);
        wdg = strsplit(params, ';','CollapseDelimiters',false);   % separe statements
        gData.staticForm = "";
        for fld = 1:numel(wdg)
            if (isempty(wdg{fld})) continue; end
            field = strsplit(wdg{fld},",", "CollapseDelimiters", false);
            if (this.fg_Left(field{1},1) == "%")  % is comment
                continue
            end
            for j=numel(field)+1:12 % normalize
                field{j} = [];
            end
            type = upper(field{1});
            field{1} = type;
            switch (type)
              case "FORM"
                  Title = this.replace(field{3},plcHolders);
                  if (~isempty(field{4})) formColor = this.checkColor(field{4},formColor); end
                  this.windowStyle = 'normal';
                  if (~isempty(field{5})) this.windowStyle = field{5}; end
                  if (strcmpi(field{6},"static")) 
                    buttons = rmfield (buttons, "fg_Cancel");
                    gData.staticForm = "S";
                  end
                  if (~isempty(field{7})) formPosition(1) = str2double(field{7}); end   % left
                  if (~isempty(field{8})) formPosition(2) = str2double(field{8}); end   % bottom
                  if (~isempty(field{9}))   % width
                    formPosition(3) = str2double(field{9});
                  end               
                  if (~isempty(field{10})) formPosition(4) = str2double(field{10}); end % height
               case "EVENT"
                   gData.Events.(field{2}).eventName = field{3};
                   gData.Events.(field{2}).handler = field{4};
               case "H"    % hidden
                   field{3} = this.replace(field{3},plcHolders);
                   field{5} = field{3}; % field{3} can be modified
               case "LINK"
                  item = strsplit(char(field{2}),'\s+','DelimiterType','RegularExpression');
                  if (numel(item) > 1) Links.(item{1}) = item{2}; end
              otherwise                   % the widgets
                  if (isempty(field{2}))  % no field name
                    field{2} = strcat(field{1},sprintf("_%i",fld));
                  end
                  if (isempty(field{3}))  % label is name if doesn't exists
                    field{3} = field{2};
                  end
                  if (isempty(field{4}))  % if doesn't exists length
                    field{4} = "20";
                    if (isfield(defLength,type))
                      field{4} = defLength.(type);
                    end
                  end
                  field{4} = str2double(field{4});
                  if strcmp(type,"T") && field{4} > 50
                      field{1} = "A";   % area text
                  end               
                  nWidget = nWidget + 1;
                  if (~strcmp(type,"C") && length(field{3}) > lLabels)
                      lLabels = length(field{3});
                  end
                  field{3} = this.replace(field{3},plcHolders);    % label
                  field{5} = this.replace(field{5},plcHolders);    % default
                  field{6} = this.replace(field{6},plcHolders);    % extra
                  field{7} = this.replace(field{7},plcHolders);    % extra 2
                  gData.data.(field{2}) = field;
            end
        end
  % Form creation
      lLabels = lLabels * charWidth;
      this.gfc = figure("Menubar","none","numbertitle","off","tag","fg_Form",...
            "visible","off",...
            "Color",formColor,"windowStyle",this.windowStyle,"name",Title);
      set(this.gfc,'CloseRequestFcn',{@fg_HandleButtons, this, handleAnswer});
      fPosition = get(this.gfc,"Position");
      sWidth = fPosition(3);
      if formPosition(3) > 0 sWidth = formPosition(3); end
      sHeight = fPosition(4);
      gData.fg_pnl = this.gfc;
      fn = fieldnames(gData.data);
      xWidgetPos = xPos+lLabels+4;
      for k=1:numel(fn)
          step = deltaY;
          name = char(fn(numel(fn)+1-k));
          wdg = gData.data.(name);
          type = wdg{1};
          if (strcmp(type,"H")) continue; end
          btnLength = buttonsDim;
          if (~isempty(wdg{4})) btnLength = wdg{4}; end
          if (strcmp(type,"B"))   % my be extra button
            if isfield(buttons,wdg{2})
              buttons.(wdg{2}){2} = wdg{3}; % change caption
              continue; 
            end
            btn = uicontrol ("style", "pushbutton", "string", wdg{3});
            set(btn,"tag",wdg{2});
            if (~isempty(wdg{5})) set(btn,"CallBack", {str2func(wdg{5}), this}); end
            this.tooltip(btn,wdg{7});
            if (isfield(gData.data,wdg{6}))
              handle = gData.data.(wdg{6}){9};
              if (strcmp(gData.data.(wdg{6}){1},"I")) handle = gData.data.(wdg{6}){10}; end
               pos = get(handle,"position");
               pos(1) = pos(1) + pos(3) + 10;
               set(btn,"position",this.setPos(pos,"width", btnLength));
            else
              set(btn,"position", [xWidgetPos yPos btnLength hWidget]);
              yPos = yPos - deltaY;
            end
            realWidth = this.checkWidth(realWidth,get(btn,"Position"));
            continue
         end
          fg_label = wdg(3);
          fg_length = charWidth * min(cell2mat(wdg(4)),txtWidth);
          lbl = uicontrol ("style","text", "position", [xPos yPos + 2 lLabels hWidget - 6], ...
              "backgroundcolor", formColor,"string", fg_label);
          wdg{8} = lbl;    % label handle
          set(lbl,'HorizontalAlignment',"right");     % lbl can be used on only one widget
%          set(lbl,"backgroundcolor", "cyan");
%          xWidgetPos = xPos+lLabels+4;
          switch (type)
            case "C"    % comment
              set(lbl,"position",[xPos yPos + 2 sWidth - 2*xPos hWidget - 6]);
              if isempty(wdg{5}) || strcmpi(wdg{5}, "after") wdg{5} = "left"; end
              if strcmpi(wdg{5}, "left") || strcmpi(wdg{5}, "right") || strcmpi(wdg{5}, "center")
                set(lbl,'HorizontalAlignment',wdg{5})
              end
              if (~isempty(wdg{7}))
                if strcmpi(wdg{7}, "error") wdg{7} = "red"; end
                set(lbl,"foregroundcolor", this.checkColor(wdg{7},[0 0 0]));
              end
              if (isfield(gData.data,wdg{6}))   % is an after comment
                handle = gData.data.(wdg{6}){9};
                if (strcmp(gData.data.(wdg{6}){1},"I")) handle = gData.data.(wdg{6}){10}; end
                pos = get(handle,"position");
                set(lbl, "position",[pos(1) + pos(3) + 10 pos(2) fg_length hWidget - 6]);
                continue
              end              
              realWidth = this.checkWidth(realWidth,get(lbl,"Position"));
            case "CKB"
              ckb = uicontrol ("style", "checkbox", "string", wdg{6},...
                "position", [xWidgetPos yPos fg_length hWidget],"backgroundcolor", formColor);
                wdg{9} = ckb;    % widget handle
                set(ckb, "Tag",name);
                set(ckb, "userdata",this.checkBoxValues);
                realWidth = this.checkWidth(realWidth,get(ckb,"Position"));
            case "L"
              [key, v] = this.getKeyValue(wdg{6});
              gData.itemsList.(name) = key;
              lst = uicontrol("style", "popupmenu","string", v, ...
                    "callback",{@fg_HandleEvents, this}, ...
                    "position", [xWidgetPos yPos fg_length hWidget]);
              wdg{9} = lst;    % widget handle
              set(lst, "Tag",name);     % lst can be used when only one widget
              realWidth = this.checkWidth(realWidth,get(lst,"Position"));
            case "R"    % radiobutton
              [key, v] = this.getKeyValue(wdg{6});
              % lst can be used when only one widget
              lst = uibuttongroup( "Tag",name,"backgroundcolor", formColor,...
                    "unit","pixels","Position",[xWidgetPos yPos fg_length * numel(v) hWidget]);
              realWidth = this.checkWidth(realWidth,get(lst,"Position"));
              key{1,numel(key)+1} = '';
              gData.itemsList.(name) = key;
              % hidden radio button
              bx = uicontrol(lst, "style", "radiobutton","string", "","Tag",name, ...
                    "Position", [2 * fg_length 1 fg_length hWidget]);
              set(bx,"userdata",sprintf("%i",numel(key)),"visible","off")             
              rdbHandlers = zeros(numel(key));
              rdbHandlers(numel(key)) = bx;
              for kRdb = 1:numel(v)              
                    bx = uicontrol(lst, "style", "radiobutton","string", v(kRdb), ...
                          "Position", [2+(kRdb-1)* fg_length 1 fg_length hWidget]);
                    set(bx,"CallBack",{@fg_HandleButtons, this, handleAnswer});
                    set(bx,"Tag",name);
                    set(bx,"userdata",sprintf("%i",kRdb))
                    rdbHandlers(kRdb) = bx;
              end
              set(lst,"userdata",rdbHandlers);
              wdg{9} = lst;    % widget handle
            case "S"    % slider
              sld = uicontrol ("style", "slider", "callback",{@fg_HandleEvents, this}, ...
                  "position", [xWidgetPos yPos fg_length hWidget]);    
              wdg{9} = sld;    % widget handle
              set(sld, "Tag",name);
              lbl = uicontrol ("style","text","backgroundcolor", formColor,...
                  "HorizontalAlignment","right",...
                  "position", [2 + xWidgetPos + fg_length yPos + 2 50 hWidget - 6]);
              range = strsplit(strtrim(strcat(wdg{6}," 0 1"))," ");
              uData.handle = lbl;
              uData.m = str2double(range(1));
              uData.M = str2double(range(2));
              uData.decimals = -floor(log10(abs(uData.M-uData.m))-2);  % 1/(10^((floor(log10(delta)-2) < 1)?-Math.floor(log10(delta)-2):0)));
              if (uData.decimals < 0) uData.decimals = 0; end
              set(sld, "userdata",uData);
              set(sld,"min",min(uData.m,uData.M));
              set(sld,"max",max(uData.m,uData.M));
              if (~isempty(wdg{6}))    % check default in range
                 if (~this.fg_Inside(str2double(wdg{5}),uData.m,uData.M))
                     sprintf("Slider default %s outside range\n",name);
                     wdg{5} = sprintf("%f",min(uData.m,uData.M));
                 end
              end
              realWidth = this.checkWidth(realWidth,get(lbl,"Position"));
            case "A"    % area text
               txt = uicontrol ("style","edit","HorizontalAlignment","left","Tag",name);            
              rows = ceil(cell2mat(wdg(4))/ txtWidth);
              step = deltaY +(rows-1) * hWidget;
              set(txt, "position", [xWidgetPos yPos fg_length hWidget * rows]);
              set(txt, "min",0,"max", rows);
              this.tooltip(txt,wdg{6});
              wdg{9} = txt;    % widget handle
              realWidth = this.checkWidth(realWidth,get(txt,"Position"));
            case {"T","I","N","NS","NF"}
              txt = uicontrol ("style","edit","callback",{@fg_HandleEvents, this});            
              set(txt, "position", [xWidgetPos yPos fg_length hWidget]);
              this.tooltip(txt,wdg{6});
              wdg{9} = txt;    % widget handle
              set(txt, "Tag",name);
              realWidth = this.checkWidth(realWidth,get(txt,"Position"));
              if (strcmp(type,"I"))  %imaginary field
                txti = uicontrol ("style","edit", "position", [xWidgetPos + 2 + fg_length yPos fg_length hWidget],...
                                  "HorizontalAlignment","right","callback",{@fg_HandleEvents, this});
                wdg{10} = txti;    % widget handle
                set(txti, "Tag",name);
                realWidth = this.checkWidth(realWidth,get(txti,"Position"));
              end
              if (isfield(numerics,type))
                 set(txt,'HorizontalAlignment',"right");
              else
                 set(txt,'HorizontalAlignment',"left");
              end
            otherwise
                sprintf("Unknown type: %s",type)
                set(lbl,"foregroundColor",[1 0 0]);
          end
          gData.data.(name) = wdg;
          yPos = yPos + step;
      end
      if formPosition(1) > -1 fPosition(1) = formPosition(1); end
      if formPosition(2) == -1 && yPos > sHeight
          fPosition(2) = fPosition(2) - (yPos - sHeight);
      end
      if formPosition(3) == -1
          fPosition(3) = realWidth+ xPos;
      else
          fPosition(3) = formPosition(3);
      end
      if formPosition(4) == -1 && yPos > sHeight
          fPosition(4) = fPosition(4) + yPos - sHeight;
      else
          fPosition(4) = yPos;
      end
      set(this.gfc,"Position",fPosition);      
      % create buttons
      guidata(this.gfc, gData);
      if (k == 1 && (strcmp(type,"L") || strcmp(type,"R")))
        if (strcmp(type,"L"))    % Radio buttons already set
          set(lst,"CallBack", {@fg_HandleButtons,this, handleAnswer});
        end
        lstAfter = get(lst,"Position");
        lstAfter(2) = lstAfter(2) - 15;
        set(lst,"Position",lstAfter);
        lblAfter = get(lbl,"Position");
        lblAfter(2) = lblAfter(2) - 15;
        set(lbl,"Position",lblAfter);        
      else
        fn = fieldnames(buttons);
        startBtn = 0.5 * (fPosition(3) - numel(fn) * (buttonsDim + 5));
        for k=1:numel(fn)
          btn = uicontrol ("string",buttons.(char(fn(k)))(2), ...
                          "position", [startBtn 10 buttonsDim hWidget]);
          set(btn,"Tag",char(fn(k)));
          set(btn,"CallBack", {@fg_HandleButtons,this, handleAnswer});
          if (strcmp(char(fn(k)),"fg_Reset")) handleReset = btn; end
          startBtn = startBtn + buttonsDim + 5;
        end
          fg_HandleButtons(handleReset,'');   % set defaults
      end
      % set Links
      fn = fieldnames(Links);
      for k=1:numel(fn)
          key = char(fn(k));
          if (~isfield(gData.data,key))
              sprintf("Link widget %s doesn't exists",key)
              continue
          end
          if (isfield(gData.data,Links.(key)))
              set(gData.data.(key){9}, "userdata", gData.data.(Links.(key)){9});
          else
              sprintf("Link widget %s doesn't exists",Links.(key))
          end
      end
      set(this.gfc,"visible","on");
      % set events
      fn = fieldnames(gData.Events);
      for k=1:numel(fn)
          key = char(fn(k));
          if (isfield(gData.data,key))
              wdg = gData.data.(key);
              cBack = get(wdg{9},"CallBack");
              if strcmp(cBack,"")
                  set(wdg{9},"CallBack", {str2func(gData.Events.(key).handler), ...
                      this,gData.Events.(key).eventName});
              end
          else
            sprintf("Event in non existent field: %s",key)  
          end
      end
    end
  end
end